import java.lang.*;
import java.util.*;

public class DataNode {
    String _node_label = null;
    String _node_data = null;
    int _node_path = 0;

    int _num_children = 0;
    int _num_parents = 0;
    Vector _child_offsets = null;
    Vector _parent_offsets = null;

    FileVector _child_offsets_on_file = null;

    /**
     * constructor.
     * 
     * @param node_label_ label for this data node
     */
    public DataNode(String node_label_) {
	_node_label = node_label_;
	_node_data = new String("");
	_child_offsets = new Vector();
	_parent_offsets = new Vector();       
    }

    /**
     * function to set the data for this data node.
     *
     * @param node_data_ String object representing the data for this node
     */
    public void setData(String node_data_) {
	_node_data = node_data_;
    }
    /** 
     * function that returns the data for this data node.
     *
     * @return String object representing the data for this data node
     */
    public String getData() {
	return _node_data;
    }
    
    /**
     * function to set the path for this data node.
     *
     * @param node_path_ integer index of the dataguide node for this data node
     */
    public void setPath(int node_path_) {
	_node_path = node_path_;
    }
    
    /**
     * function that returns the path for this data node.
     * 
     * @return integer index of the dataguide node for this data node
     */
    public int getPath() {
	return _node_path;
    }

    /**
     * function to add a parent for this data node.
     *
     * @param parent_offset_ offset of the parent data node
     */
    public void addParentOffset(long parent_offset_) {
	_parent_offsets.addElement(new Long(parent_offset_));
	_num_parents++;
    }

    /**
     * function that returns all parents of this data node.
     *
     * @return Vector of all parent offsets
     */
    public Vector getParentOffsets() {
	return _parent_offsets;
    }

    /**
     * function to get the offset of the first child of this data node.
     *
     * @return offset of the first child
     */
    public long getFirstChildOffset() {
	if(_child_offsets.size() == 0) return -1;
	else return ((Long)(_child_offsets.elementAt(0))).longValue();
    }

    /**
     * function to add a new child to this data node.
     *
     * @param child_offset_ offset of new child
     */
    public void addChildOffset(long child_offset_) {
	_child_offsets.addElement(new Long(child_offset_));
	_num_children++;
    }

    /**
     * function to get all children for this data node.
     *
     * @return FileVector object pointing to all the children of this data node
     */
    public FileVector getChildOffsetsOnFile() {
	return _child_offsets_on_file;
    }

    /**
     * function to get all children for this data node.
     *
     * @return Vector object containing all the children of his data node.
     */
    public Vector getChildOffsets() {
	return _child_offsets;
    }

    /**
     * function to set a label for this data node.
     *
     * @param label_ String object representing the label for this data node
     */
    public void setLabel(String label_) {
	_node_label = label_;
    }
    
    /**
     * function that returns the label of this data node.
     * 
     * @return String object representing the label of this data node
     */
    public String getLabel() {
	return _node_label;
    }
    
    /**
     * function that returns true if this data node has no children.
     *
     * @return true if this data node has no children
     *         false otherwise
     */
    public boolean isLeaf() {
	if(_child_offsets.size() == 0) return true;
	else return false;
    }  
    
    /**
     * function to add a child.
     *
     * @param k position of child among its siblings
     * @param child_offset_ offset of child
     */
    public void addChildOffset(int k, long child_offset_) {
	_child_offsets.add(k, new Long(child_offset_));
	_num_children++;
    }
    /** 
     * function to delete all children of this data node.
     */
    public void removeAllChildren() {
	_child_offsets = new Vector();
	_num_children = 0;
    }

    /** 
     * function to delete all parents of this data node.
     */
    public void removeParents() {
	_parent_offsets = new Vector();
	_num_parents = 0;
    }
    /**
     * function to set the location of this data node's children.
     *
     * @param file_name_ name of the file containing the children offsets 
     * @param offset_ offset in file where the children offsets are stored
     * @param num_children_ number of children for this data node
     */
    public void setChildrenLocation(String file_name_, long offset_, int num_children_) {
	_child_offsets_on_file = new FileVector(file_name_, offset_);
	_num_children = num_children_;	
	_child_offsets_on_file.setSize(_num_children);
    }
}

