import java.lang.*;
import java.util.*;

public class DataguideNode {
    private String _dg_node_id = null; // unique id, generated using a random number generator
    private String _dg_node_label = null; // label on this node
    private String _dg_node_path = null; // path for this node
    private int _dg_node_index = -1;
    private int _distance_from_root = 0;
    //private HashSet _data_nodes = null; // data nodes represented by this dataguide node
    private FileVector _file_vector = null;
    private String  _index_directory = null;

    /**
     * constructor.
     *
     * @param dg_node_index_ index of dg node
     * @param index_directory_ directory to store files representing this dg node
     * @param dg_node_id_ identifier for this dg node
     * @param dg_node_label_ label for this dg node
     */

    DataguideNode(int dg_node_index_, String index_directory_, String dg_node_id_, String dg_node_label_) {
	_dg_node_index = dg_node_index_;
	_dg_node_id = dg_node_id_;
	_dg_node_label = dg_node_label_;
	_index_directory = index_directory_;
	//_data_nodes = data_node_set_;
	
	if(MyFrame._debug_mode > 1) {
	    System.out.println("Creating dataguide node");
	}

	if(MyFrame._debug_mode > 5) {
	    System.out.println("ID = " + _dg_node_id); 
	}

	if(_dg_node_path != null) {
	    _file_vector = new FileVector(_index_directory + _dg_node_index +".dgnode", 0);
	}
    }

    /** 
     * function that flushes the data nodes of this dg node to disk.
     */
    public void flush() {
	_file_vector.flush();
    }
    /**
     * function that returns the number of data nodes for this dg node.
     *
     * @return number of data nodes for this dg node
     */
    public int getNumDataNodes() {
	return _file_vector.size();
    }

    /**
     * function that sets the number of data nodes for this dg node.
     *
     * @param num_data_nodes_ number of data nodes for this dg node
     */
    public void setNumDataNodes(int num_data_nodes_) {
	_file_vector.setSize(num_data_nodes_);
    }

    /**
     * function to add a data node to this dg node.
     *
     * @param curr_offset_ offset of the data node to be added
     */
    public void addDataNode(Long curr_offset_) {
	if(_dg_node_path == null) { 
	    System.out.println("Set path before calling addDataNode");
	    System.exit(0);
	}

	if(_file_vector == null)  {
	    _file_vector = new FileVector(_index_directory + _dg_node_index +".dgnode", 0);
	}
	_file_vector.addElement(curr_offset_);
    }

    /**
     * function to returns a vector of all data nodes for this dg node.
     *
     * @return FileVector object having a pointer to the offsets of all data nodes for this dg node
     */
    public FileVector getNodeSet() {
	return _file_vector;
    }

    /**
     * function that returns the identifier for this dg node.
     *
     * @return String object representing the id for this dg node
     */
    public String getId() {
	return _dg_node_id;
    }

    /**
     * function that returns the label on this dg node.
     *
     * @return String object representing the label on this dg node
     */
    public String getLabel() {
	return _dg_node_label;
    }	

    /**
     * function to set the path for this dg node.
     *
     * @param dg_node_path_ path for this dg node
     * @param distance_ depth for this dg node
     */
    public void setPath(String dg_node_path_, int distance_) {
	_dg_node_path = dg_node_path_;
	_distance_from_root = distance_;
	if(_file_vector == null)  {
	    _file_vector = new FileVector(_index_directory + _dg_node_index +".dgnode", 0);
	}
    }

    /**
     * function that returns the index of the current dg node.
     *
     * @return index of current dg node
     */
    public int getDGIndex() {
	return _dg_node_index;
    }

    /**
     * function that returns the path for this dg node.
     *
     * @return String object representing the path for this dg node
     */
    public String getPath() {
	return _dg_node_path;
    }

    /**
     * function that returns the depth for this dg node.
     *
     * @return depth for this dg node
     */
    public int getDistance() {
	return _distance_from_root;
    }
}
