import java.lang.*;
import java.util.*;
import java.io.*;

public class FileVector {
    private DataOutputStream _data_out_stream = null;
    static private int BUFFER_SIZE = 32768;
    static private int LONG_SIZE = 8;
    private int _size = 0;
    private RandomAccessFile _random_access_file = null;
    private String _data_file_name = null;
    private long _starting_offset = 0;
    private DataInputStream _data_in_stream = null;
    private RandomAccessFile _random_access_out = null;

    /**
     * constructor.
     *
     * @param data_file_name_ name of file to write the file vector
     * @param offset_ offset in file where the file vector begins
     */
    FileVector(String data_file_name_, long offset_) {
	_data_file_name = data_file_name_;
	_starting_offset = offset_;
    }

    protected void finalize() throws Throwable {
	try {
	    if(_data_out_stream != null) {
		_data_out_stream.close();
		_data_out_stream = null;
	    }
	    if(_random_access_out != null) {
		_random_access_out.close();
		_random_access_out = null;
	    }
	    if(_random_access_file != null) {
		_random_access_file.close();
		_random_access_file = null;
	    }
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file close " + ioe.toString());
	    System.exit(0);
	}
    }

    /**
     * function to initialize a sequential scan of the file vector.
     */     
    public void beginScan() {
	if(_data_out_stream != null) {
	    try {
		_data_out_stream.close();
		_data_out_stream = null;
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file close " + ioe.toString());
		System.exit(0);
	    }
	}
	if(_random_access_out != null) {
	    try {
		_random_access_out.close();
		_random_access_out = null;
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file close " + ioe.toString());
		System.exit(0);
	    }
	}

	if(_data_in_stream == null) {
	    try {
		_data_in_stream = new DataInputStream(new BufferedInputStream(new FileInputStream(_data_file_name), BUFFER_SIZE));
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	}
    }

    /**
     * function to return the next item in the file vector.
     *
     * @return next Long object in the file vector
     */
    public Long nextElement() {
	long value = -1;
	try {
	    value = _data_in_stream.readLong();
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file write " + ioe.toString());
	    System.exit(0);
	}
	return new Long(value);
    }

    /**
     * function to end a sequential scan of the file vector.
     */
    public void endScan() {
	try {
	    if(_data_in_stream != null) _data_in_stream.close();
	    _data_in_stream = null;
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file flush " + ioe.toString());
	    System.exit(0);
	}
    }

    /** 
     * function to flush the file vector to disk.
     */
    public void flush() {
	try {
	    if(_data_out_stream != null) _data_out_stream.close();
	    _data_out_stream = null;
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file flush " + ioe.toString());
	    System.exit(0);
	}
    }

    /**
     * function that returns the size of the file vector.
     *
     * @return size of file vector
     */
    public int size() {
	return _size;
    }

    /**
     * function to set the size of the file vector.
     *
     * @param size_ new size of file vector
     */
    public void setSize(int size_) {
	_size = size_;
    }

    /**
     * function to add a new item to the file vector.
     *
     * @param curr_offset_ new offset to be added to the file vector
     */
    public void addElement(Long curr_offset_) {
	if(_data_out_stream == null) {
	    try {
		_data_out_stream = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(_data_file_name), BUFFER_SIZE));
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	}

	    
	try {
	    _data_out_stream.writeLong(curr_offset_.longValue());
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file write " + ioe.toString());
	    System.exit(0);
	}
	_size++;
    }

    /**
     * function to add a new item to the file vector.
     *
     * @param curr_offset_ new offset to be added to the file vector
     */
    public Long elementAtRandom(int index_) {
	try {
	    _random_access_out = new RandomAccessFile(_data_file_name, "r");
	    _random_access_out.seek(_starting_offset+LONG_SIZE*index_);
	}
	catch(FileNotFoundException fnfe) {
	    System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
	    System.exit(0);
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}
	
	long value = -1;
	try {
	    value = _random_access_out.readLong();
	    _random_access_out.close();
	    _random_access_out = null;
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file read " + ioe.toString());
	    System.exit(0);
	}
	return new Long(value);
    }

    public void addElementRandom(Long curr_offset_) {
	try {
	    _random_access_out = new RandomAccessFile(_data_file_name, "rw");
	    System.out.println("Seeking to offset "+(_starting_offset+LONG_SIZE*_size));
	    _random_access_out.seek(_starting_offset+LONG_SIZE*_size);
	}
	catch(FileNotFoundException fnfe) {
	    System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
	    System.exit(0);
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file creation " + ioe.toString());
	    System.exit(0);
	}
	
	try {
	    _random_access_out.writeLong(curr_offset_.longValue());
	    _random_access_out.close();
	    _random_access_out = null;
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file write " + ioe.toString());
	    System.exit(0);
	}
	_size++;
    }

    /**
     * function that returns the object at the location specified by the argument.
     *
     * @param index_ index of object
     * @return the object
     */
    public Long elementAt(int index_) {
	if(_data_out_stream != null) {
	    try {
		_data_out_stream.close();
		_data_out_stream = null;
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file close " + ioe.toString());
		System.exit(0);
	    }
	}
	if(_random_access_out != null) {
	    try {
		_random_access_out.close();
		_random_access_out = null;
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file close " + ioe.toString());
		System.exit(0);
	    }
	}
	if(_random_access_file == null) {
	    
	    try {
		_random_access_file = new RandomAccessFile(_data_file_name, "r");
	    }
	    catch(FileNotFoundException fnfe) {
		System.out.println("FileNotFoundException during data file creation " + fnfe.toString());
		System.exit(0);
	    }
	    catch(IOException ioe) {
		System.out.println("IOException during data file creation " + ioe.toString());
		System.exit(0);
	    }
	}
	
	long value = -1, offset = -1;
	try {
	    offset = _starting_offset + LONG_SIZE*index_;
	    _random_access_file.seek(offset);	
	    value = _random_access_file.readLong();
	}
	catch(IOException ioe) {
	    System.out.println("IOException during data file seek offset = " + offset + ", start = "+_starting_offset);
	    System.exit(0);
	}
	return new Long(value);
    }
}

