import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

public class MyLabelBox {
    private boolean _displayed_all_text = false;
    static private double TEXT_CORRECTION_FACTOR = VDataguide.TEXT_CORRECTION_FACTOR;
    static private final double ABS_PEN_WIDTH = VDataguide.ABS_PEN_WIDTH;
    private double _label_box_height = 0;
    /**
     * constructor.
     *
     * @param my_canvas_ canvas on which to draw the data node
     * @param node_ Jazz object 
     * @param sticky_node_ not being used
     * @param top_left_x_ x-coord of top-left corner of bounding box
     * @param top_left_y_ y-coord of top-left corner of bounding box
     * @param bbox_width_ width of bounding box
     * @param bbox_height_ height of bounding box
     * @param label_ String object representing the text to be displayed
     * @param avg_char_width_ average width of a character at the current magnification level
     * @param avg_char_height_ average height of a character at the current magnification level
     * @param font_size_ size of the font that is readable at the current magnification level
     * @param camera_mag_ current magnification level
     */
    MyLabelBox(ZCanvas my_canvas_, VDataguide node_, ZVisualLeaf sticky_node_, 
	       double top_left_x_, double top_left_y_, double bbox_width_, double bbox_height_,
	       String label_, double avg_char_width_, double avg_char_height_, 
	       int font_size_, double camera_mag_) {
	// create bounding box
	ZPolyline text_bbox = new ZPolyline();
 	text_bbox.add(top_left_x_, top_left_y_);
	text_bbox.add(top_left_x_+bbox_width_, top_left_y_);
	text_bbox.add(top_left_x_+bbox_width_, top_left_y_+bbox_height_);
	text_bbox.add(top_left_x_, top_left_y_+bbox_height_);
	text_bbox.add(top_left_x_, top_left_y_);
	text_bbox.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	node_.addVisualComponent(text_bbox);

	//ZPolyline text_bbox = new ZRectangle(top_left_x_, top_left_y_, bbox_width_, bbox_height_);	

	// compute num_chars_per_line and num_lines that can fit into the text box
	int num_chars_per_line = (int)((bbox_width_*TEXT_CORRECTION_FACTOR)/avg_char_width_);
	if(MyFrame._debug_mode > 7) {
	    System.out.println("MyLabelBox: No. of characters per line="+num_chars_per_line);
	}
	// Use the computed parameters (num_chars_per_line, num_lines) to break the text 
	// into lines
	int label_len = label_.length();
	StringBuffer new_text_buffer = new StringBuffer(" ");		
	int i=0;
	for(i=0; i<(num_chars_per_line-1) && i<label_len; i++) {
	    new_text_buffer.append(label_.charAt(i));
	}		
	if(i == label_len) {_displayed_all_text = true;}

	String new_label = new_text_buffer.toString();
	Font label_font = new Font(null, Font.PLAIN, font_size_);
	ZLabel label_visual_obj = new ZLabel(new_label, label_font);
	label_visual_obj.setFont(label_font);
	if(_displayed_all_text)
	    label_visual_obj.setPenColor(Color.red);
	else
	    label_visual_obj.setPenColor(Color.black);
	Point2D location=new Point2D.Double(top_left_x_, top_left_y_+bbox_height_/2.0f-avg_char_height_/2.0f);
	label_visual_obj.setTranslation(location);
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Location of child label in normal node coords=("+location.getX()+","+location.getY()+")");
	}
	//my_canvas_.getCamera().localToCamera(location, node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of child label in camera coords=("+location.getX()+","+location.getY()+")");
	//}
	//my_canvas_.getCamera().cameraToLocal(location, sticky_node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of child label in sticky node coords=("+location.getX()+","+location.getY()+")");
	//}
	//label_visual_obj.setTranslation(location);
	//sticky_node_.addVisualComponent(label_visual_obj);
	node_.addVisualComponent(label_visual_obj);
	
	_label_box_height = label_visual_obj.getBounds().getHeight();///camera_mag_;	    
    }

    /**
     * function that returns the height of the label box.
     * 
     * @return double value of the label box's height 
     */

    public double getLabelBoxHeight() {
	return _label_box_height;
    }

    /** 
     * function to check if all of the text was displayed.
     * 
     * @return boolean value indicating if all text was displayed
     */

    public boolean allTextDisplayed() {
	return _displayed_all_text;
    }
}
