import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

public class MyTextBox {
    private boolean _displayed_all_text = false;
    private double _text_box_height = 0;

    static public final int READABLE_FONT_SIZE = VDataguide.READABLE_FONT_SIZE;
    static public final double TEXT_CORRECTION_FACTOR = VDataguide.TEXT_CORRECTION_FACTOR;
    static public final int MAX_FONT_SIZE = VDataguide.MAX_FONT_SIZE;
    static public final int MIN_FONT_SIZE = VDataguide.MIN_FONT_SIZE;

    /**
     * constructor.
     *
     * @param my_canvas_ canvas on which to draw the data node
     * @param node_ Jazz object 
     * @param sticky_node_ not being used
     * @param top_left_x_ x-coord of top-left corner of bounding box
     * @param top_left_y_ y-coord of top-left corner of bounding box
     * @param bbox_width_ width of bounding box
     * @param bbox_height_ height of bounding box
     * @param text_ String object representing the text to be displayed
     * @param camera_mag_ current magnification level
     */
    MyTextBox(ZCanvas my_canvas_, VDataguide  node_, ZVisualLeaf sticky_node_, 
	      double top_left_x_, double top_left_y_, 
	      double bbox_width_, double bbox_height_,
	      String text_, double camera_mag_) {
	// create sample text
	String []sample_text_lines = {"HLIK S-9V}", "WGglMyBm#/", "SPEAKERSPE"};
	int num_sample_lines = sample_text_lines.length;
	String sample_text = "";
	int max_line_len = 0;
	for(int i=0; i<num_sample_lines; i++) {
	    sample_text_lines[i].trim();
	    int curr_line_len = sample_text_lines[i].length();
	    if(curr_line_len > max_line_len) max_line_len = curr_line_len;
	    sample_text = sample_text + sample_text_lines[i];
	    if(i != (num_sample_lines-1)) sample_text = sample_text + "\n";
	}

	// create ZText object corresponding to sample text
	double font_size_d = (READABLE_FONT_SIZE/camera_mag_);
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Font size = " + font_size_d);   
	}
	int font_size = (int)(Math.ceil(font_size_d));
	if(font_size > MAX_FONT_SIZE) font_size = MAX_FONT_SIZE;
	else if(font_size < MIN_FONT_SIZE) font_size = MIN_FONT_SIZE;

	if(MyFrame._debug_mode > 8) {
	    System.out.println("Inside MyTextBox constructor");
	    System.out.println("Camera Mag = " + camera_mag_);   
	    System.out.println("Adjusted Font size = " + font_size);   
	}

	// find coordinates of bounding box	
	Font sample_text_font = new Font(null, Font.PLAIN, font_size);
	ZText text_visual_obj = new ZText(sample_text);
	text_visual_obj.setFont(sample_text_font);
	text_visual_obj.setPenColor(Color.black);
	text_visual_obj.setTranslation(top_left_x_, top_left_y_);
	
	// find average character width and height
	ZBounds sample_text_bounds = text_visual_obj.getBounds();
	double sample_text_width = sample_text_bounds.getWidth();
	double sample_text_height = sample_text_bounds.getHeight();
	double avg_char_width = sample_text_width/max_line_len;
	double avg_char_height = sample_text_height/num_sample_lines;

	// compute num_chars_per_line and num_lines that can fit into the text box
	//int num_chars_per_line = (int)((bbox_width_*camera_mag_*TEXT_CORRECTION_FACTOR)/avg_char_width);
	int num_chars_per_line = (int)((bbox_width_*TEXT_CORRECTION_FACTOR)/avg_char_width);
	//int max_num_lines = (int)((bbox_height_*camera_mag_*TEXT_CORRECTION_FACTOR)/avg_char_height);
	int max_num_lines = (int)((bbox_height_*TEXT_CORRECTION_FACTOR)/avg_char_height);
	if(MyFrame._debug_mode > 7) {
	    System.out.println("MyTextBox: Max. number of lines="+max_num_lines);
	    System.out.println("MyTextBox: No. of characters per line="+num_chars_per_line);
	}
	
	// Use the computed parameters (num_chars_per_line, max_num_lines) to break the text 
	// into lines
	String old_text = text_; 
	StringTokenizer st = new StringTokenizer(old_text);
	int rem_chars_in_line = num_chars_per_line;
	StringBuffer new_text_buffer = new StringBuffer("");		
	int line_num = 0;
	String curr_word=null;
	int curr_word_len=0;
	boolean get_next_token=true;
	
	while(line_num<max_num_lines && (st.hasMoreTokens() || get_next_token == false)) {
	    // get next token
	    if(get_next_token) {
		curr_word = st.nextToken();
		curr_word.trim();
		curr_word_len = curr_word.length();
		get_next_token = false;
	    }
	    
	    // if possible, append the word to the current line
	    if(MyFrame._debug_mode > 8) {
		System.out.println("line="+(line_num+1)+",word=\""+curr_word+"\""+",rem="+rem_chars_in_line+",len="+curr_word_len);
	    }
	    if(curr_word_len <= (rem_chars_in_line-1)) {
		new_text_buffer.append(" ");
		new_text_buffer.append(curr_word);
		get_next_token = true;
		rem_chars_in_line = rem_chars_in_line-(curr_word_len+1);
		if(rem_chars_in_line <= 1) {				
		    // go to the next line
		    if(line_num != (max_num_lines-1)) new_text_buffer.append("\n");
		    line_num++;
		    rem_chars_in_line = num_chars_per_line;
		}
	    }
	    // else go to the next line
	    else {
		if(rem_chars_in_line == num_chars_per_line) {
		    // break the word 
		    new_text_buffer.append(" ");
		    for(int i=0; i<(rem_chars_in_line-1); i++) {
			new_text_buffer.append(curr_word.charAt(i));
		    }
		    break;
		}
		if(line_num != (max_num_lines-1)) new_text_buffer.append("\n");
		line_num++;
		rem_chars_in_line = num_chars_per_line;
	    }
	}
	String new_text = new_text_buffer.toString();
	if(MyFrame._debug_mode > 6) {
	    System.out.println("New Text["+num_chars_per_line+","+max_num_lines+"]:");
	    System.out.println(new_text);
	}
	text_visual_obj = new ZText(new_text);
	text_visual_obj.setFont(sample_text_font);
	// if all the text has been displayed, set font color to red
	if(st.hasMoreTokens() || get_next_token == false) {
	    text_visual_obj.setPenColor(Color.black);	    
	}
	else {
	    _displayed_all_text = true;
	    text_visual_obj.setPenColor(Color.red);
	}
	Point2D location=new Point2D.Double(top_left_x_, top_left_y_);
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Location of data node text in normal node coords=("+location.getX()+","+location.getY()+")");
	}
	//my_canvas_.getCamera().localToCamera(location, node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of data node text in camera coords=("+location.getX()+","+location.getY()+")");
	//}
	//my_canvas_.getCamera().cameraToLocal(location, sticky_node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of data node text in sticky node coords=("+location.getX()+","+location.getY()+")");
	//}
	text_visual_obj.setTranslation(location);
	//sticky_node_.addVisualComponent(text_visual_obj);
	node_.addVisualComponent(text_visual_obj);

	_text_box_height = text_visual_obj.getBounds().getHeight();///camera_mag_;

	if(MyFrame._debug_mode > 8) {
	    System.out.println("Exiting MyTextBox constructor........");
	}
    }

    /**
     * function that returns the height of the text box.
     * 
     * @return double value of the text box's height 
     */
    public double getTextBoxHeight() {
	return _text_box_height;
    }

    /** 
     * function to check if all of the text was displayed.
     * 
     * @return boolean value indicating if all text was displayed
     */
    public boolean allTextDisplayed() {
	return _displayed_all_text;
    }
}
