import java.lang.*;
import java.util.*;
import java.awt.geom.*;
import java.awt.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;
import javax.swing.*;
import com.stevesoft.pat.*;

public class VDataguide extends ZVisualLeaf {
    static public  final int    DG_GROUP_SWITCH_CAMERA_MAGNIFICATION = 15;
    static public  final int    CAMERA_MAG_SCALE_FACTOR = 10;
    static public  final int    MAX_FONT_SIZE = 40;
    static public  final int    MIN_FONT_SIZE = 1;

    static public  final double TEXT_CORRECTION_FACTOR = 1.0f;
    static public  final int    READABLE_FONT_SIZE = 10;
    static public  final double MAG_CHANGE_IGNORE = 0.01;

    static public  final double NODE_HEIGHT = 40.0;
    static public  final double NODE_WIDTH  = 80.0;
    static private final double SIZE_BOX_NODE_WIDTH_RATIO = 6.0;
    static private final double SIZE_BOX_NODE_HEIGHT_RATIO = 3.0;
    static private final double ENLARGE_FACTOR = 1.5;
    static private final double SHRINK_FACTOR = 1.5;
    static private final double SCALE_TRANSFORM_FACTOR = 1.2;

    static private final double INTER_NODE_SPACING = 40.0;
    static public  final double INTER_NODE_SPACING_X = (double) 3;
    static public  final double INTER_NODE_SPACING_Y = (double) 3;
    static public  final double ABS_PEN_WIDTH = .1f;

    private Dataguide _dataguide = null;
    private DataGraph _datagraph = null;
    private ZCanvas _my_canvas = null;
    private ZVisualLeaf _sticky_node = null;
    private ZStickyGroup _sticky_group = null;
    private TimerThread _timer_thread = null;

    private ArrayList _dg_nodes = null; // ArrayList<DataguideNode>
    private ArrayList _dg_nodegroupsize = null;  // ArrayList<Integer>
    private ArrayList _dg_nodelistpos = null; // ArrayList<Integer>
    private ArrayList _dg_rectangles = null; // ArrayList<Rectangle2D>
    private ArrayList _dg_zrectangles = null; // ArrayList<ZRectangle>
    private ArrayList _toggled_nodes = null; 
    private ArrayList _data_boxes = null; // ArrayList<ZPolyline>
    private ArrayList _prev_boxes = null; // ArrayList<ZRectangle>
    private ArrayList _next_boxes = null; // ArrayList<ZRectangle>
    private ArrayList _plus_boxes = null; // ArrayList<ZRectangle>
    private ArrayList _star_boxes = null; // ArrayList<ZRectangle>
    private ArrayList _query_boxes = null; // ArrayList<ZRectangle>
    private ArrayList _minus_boxes = null; // ArrayList<ZRectangle>

    private Vector _visible_node_indices = null; /* needed for checking if the user has panned */
    private int _num_visible_dg_nodes = 0; /* '' */
    private double _curr_mag_level = -1;

    private Hashtable _bbox_nodeoffset_map = null;
    private Hashtable _dgnode_rect_map = null;
    private long _clicked_offset = -1;
    private boolean _click_active = false;
    private boolean _repaint_flag = true;
    
    private Hashtable _dgnodeindex_histogram = new Hashtable();
    private Vector _colors = null;
    static private int _counter = 0;

    private RunningQuery _running_query = null;

    /**
     * constructor.
     *
     * @param my_canvas_ Canvas on which to draw the Jazz object
     * @param dataguide_ dataguide for the XML file
     */
    public VDataguide(ZCanvas my_canvas_, Dataguide dataguide_) {
	this.setPalette();

	_dataguide = dataguide_;
	_dataguide.flush();

	_datagraph = _dataguide.getDataGraph();
	_my_canvas = my_canvas_;
	_repaint_flag = true;
	
	// create rectangles for dataguide nodes
	_dg_rectangles = new ArrayList();
	_dg_nodes = new ArrayList();
	_dg_nodelistpos = new ArrayList();
	_dg_nodegroupsize = new ArrayList();
	_dgnode_rect_map = new Hashtable();	
	_dg_zrectangles = new ArrayList();
	_toggled_nodes = new ArrayList();

	int curr_depth = 0;
	
	int num_dg_nodes = dataguide_.getNumNodes();
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = dataguide_.getDataguideNode(i);
	    _dg_nodes.add(curr_dg_node);
	    _dg_nodelistpos.add(new Integer(0));
	    _dg_nodegroupsize.add(new Integer(VGroupNode.MAX_SAMPLE_SIZE));
	    Rectangle2D curr_dg_rect = new Rectangle2D.Double(0, 0, NODE_WIDTH, NODE_HEIGHT); // dummy value for left boundary
	    _dg_rectangles.add(curr_dg_rect);
	    ZRectangle curr_dg_zrect = new ZRectangle(curr_dg_rect.getX(), curr_dg_rect.getY(),
						      curr_dg_rect.getWidth(), curr_dg_rect.getHeight());
	    curr_dg_zrect.setPenColor(Color.black);
	    Color c= new Color(204, 255, 255 ); // pastel blue
	    curr_dg_zrect.setFillColor(c);
	    
	    _dg_zrectangles.add(curr_dg_zrect);
	    _toggled_nodes.add(new Boolean(false));
	    _dgnode_rect_map.put(curr_dg_node, curr_dg_rect);
	}
	/*
	Vector dg_nodes_at_depth = dataguide_.getNodesAtDepth(curr_depth); 
	int num_dg_nodes_at_depth = dg_nodes_at_depth.size();
	while(num_dg_nodes_at_depth > 0) {
	    for(int i=0; i<num_dg_nodes_at_depth; i++) {
		DataguideNode curr_dg_node = (DataguideNode)(dg_nodes_at_depth.elementAt(i));
		_dg_nodes.add(curr_dg_node);
		_dg_nodelistpos.add(new Integer(0));
		_dg_nodegroupsize.add(new Integer(VGroupNode.MAX_SAMPLE_SIZE));
		Rectangle2D curr_dg_rect = new Rectangle2D.Double(0, 0, NODE_WIDTH, NODE_HEIGHT); // dummy value for left boundary
		_dg_rectangles.add(curr_dg_rect);
		ZRectangle curr_dg_zrect = new ZRectangle(curr_dg_rect.getX(), curr_dg_rect.getY(),
							  curr_dg_rect.getWidth(), curr_dg_rect.getHeight());
		curr_dg_zrect.setPenColor(Color.black);
		Color c= new Color(204, 255, 255 ); // pastel blue
		curr_dg_zrect.setFillColor(c);
		
		_dg_zrectangles.add(curr_dg_zrect);
		_toggled_nodes.add(new Boolean(false));
		_dgnode_rect_map.put(curr_dg_node, curr_dg_rect);
	    }
	    curr_depth = curr_depth+1;
	    dg_nodes_at_depth = dataguide_.getNodesAtDepth(curr_depth);
	    num_dg_nodes_at_depth = dg_nodes_at_depth.size();
	}
	*/

	updateDGNodePositions();

	ZBounds global_bounds = new ZBounds();
	num_dg_nodes = _dg_nodes.size();
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(i));
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));
	    global_bounds.add(curr_dg_rect);
	}

	_dg_nodes.trimToSize();
	_dg_nodelistpos.trimToSize();
	_dg_nodegroupsize.trimToSize();
	_dg_rectangles.trimToSize();

	// create ZRectangle objects for dataguide nodes
	_prev_boxes = new ArrayList();
	_next_boxes = new ArrayList();
	_plus_boxes = new ArrayList();
	_star_boxes = new ArrayList();
	_query_boxes = new ArrayList();
	_minus_boxes = new ArrayList();
	for(int i=0; i<num_dg_nodes; i++) {
	    ZRectangle curr_dg_zrect = (ZRectangle)(_dg_zrectangles.get(i));	
	    _prev_boxes.add(new ZRectangle());
	    _next_boxes.add(new ZRectangle());
	    _plus_boxes.add(new ZRectangle());
	    _star_boxes.add(new ZRectangle());
	    _query_boxes.add(new ZRectangle());
	    _minus_boxes.add(new ZRectangle());
	    super.addVisualComponent(curr_dg_zrect);
	}

	global_bounds.transform(AffineTransform.getScaleInstance(SCALE_TRANSFORM_FACTOR, SCALE_TRANSFORM_FACTOR));
	
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Global Bounds. (x,y)=("+global_bounds.getBounds().getX()+","+global_bounds.getBounds().getY()+")");
	    System.out.println("Global Bounds. (w,h)=("+global_bounds.getBounds().getWidth()+","+global_bounds.getBounds().getHeight()+")");
	}

	// center the viewing window on the canvas
	_my_canvas.getCamera().center(global_bounds, 0, _my_canvas.getDrawingSurface());
	_my_canvas.getLayer().addChild(this);
	_sticky_node = new ZVisualLeaf();
	_sticky_group = new ZStickyGroup(_my_canvas.getCamera(), _sticky_node);
	_sticky_group.setConstraintType(ZStickyGroup.STICKYZ);
	_my_canvas.getLayer().addChild(_sticky_group);

	_timer_thread = new TimerThread(this);
	_timer_thread.start();

	_data_boxes = new ArrayList();

	_bbox_nodeoffset_map = new Hashtable();
	NodeClickListener click_handler = new NodeClickListener();
	this.addMouseListener((ZMouseListener)click_handler);			
    }

    /**
     * function to find the x-coord of the top-left corner of each dataguide node.
     *
     * @param node_ node whose x-position is to be determined
     * @param frontier_x_ current position of the global frontier
     * @return new position of the global frontier
     */
    // works only for trees
    private double findNodeTopX(DataguideNode node_, double frontier_x_) {
	// remember the frontier
	double frontier_begin = frontier_x_;
	Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(node_));
	double curr_node_width = curr_dg_rect.getWidth(); 
	
	Vector children = _dataguide.getOutEdges(node_);
	if(children.size() == 0) {
	    curr_dg_rect.setRect(frontier_begin, curr_dg_rect.getY(), curr_dg_rect.getWidth(), curr_dg_rect.getHeight());
	    return (frontier_begin + curr_node_width + INTER_NODE_SPACING);
	}
	else {
	    double frontier_end = frontier_x_;
	    Enumeration e = children.elements();
	    while(e.hasMoreElements()) {
		DataguideNode child_node = (DataguideNode)(e.nextElement());
		frontier_end = findNodeTopX(child_node, frontier_end);		
	    }
	    double subtree_width = frontier_end-INTER_NODE_SPACING-frontier_begin;
	    if(subtree_width >= curr_node_width) {
		curr_dg_rect.setRect(frontier_begin+subtree_width/2-curr_node_width/2, curr_dg_rect.getY(), 
				     curr_dg_rect.getWidth(), curr_dg_rect.getHeight());
	    }
	    else {
		curr_dg_rect.setRect(frontier_begin, curr_dg_rect.getY(), 
				     curr_dg_rect.getWidth(), curr_dg_rect.getHeight());	
		frontier_end = frontier_begin + curr_node_width + INTER_NODE_SPACING;
	    }	
	    return frontier_end;
	}
    }    
    
    private void setPalette() {
	_colors = new Vector();
	/*
	_colors.addElement(new Color(255, 255, 204)); // dirty yellow
	_colors.addElement(new Color(255, 204, 204));
	_colors.addElement(new Color(204, 204, 255));
	*/
	
	_colors.addElement(new Color(255, 204, 255));
	_colors.addElement(new Color(204, 255, 204 )); // pastel green
	_colors.addElement(new Color(153, 204, 255));
	//_colors.addElement(new Color(51, 153, 255)); // dark blue
	_colors.addElement(new Color(255, 51, 255));
	//_colors.addElement(new Color(51, 255, 255)); // flourescent blue
	//_colors.addElement(new Color(255, 255, 51)); // yellow
	//_colors.addElement(new Color(102, 0, 153)); // darker blue
	//_colors.addElement(new Color(0, 0, 102)); // darkest blue
	_colors.addElement(new Color(255, 255, 102));	
    }

    private int getFontSize(double camera_mag_) {
	double font_size_d = READABLE_FONT_SIZE/camera_mag_;
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Font size = " + font_size_d);   
	}
	int font_size = (int) (Math.ceil(font_size_d));	
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Font Size="+font_size);
	}
	if(font_size > MAX_FONT_SIZE) font_size = MAX_FONT_SIZE;
	else if(font_size < MIN_FONT_SIZE) font_size = MIN_FONT_SIZE;
	
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Adjusted Font Size="+font_size);
	}
	
	return font_size;
    }

    private void drawQueryBox(int dg_node_id_, Rectangle2D curr_bbox_, double camera_mag_) {
	int font_size = getFontSize(camera_mag_);
	Font query_label_font = new Font(null, Font.PLAIN, font_size); 
	ZLabel query_label = null;
	
	if(_running_query == null || _running_query.getNode() != dg_node_id_)
	    query_label = new ZLabel(new String("q"), query_label_font);
	else	
	    query_label = new ZLabel(new String("s"), query_label_font);
	ZBounds query_label_bounds = query_label.getBounds();
	double query_label_width = query_label_bounds.getWidth();
	double query_label_height = query_label_bounds.getHeight();

	double query_box_width = query_label_width*3;
	double query_box_height = query_label_height; 
	double query_box_top_left_x = curr_bbox_.getX()+curr_bbox_.getWidth()*1.0f/5.0f-query_box_width/2.0f;
	double query_box_top_left_y = curr_bbox_.getY()+curr_bbox_.getHeight();
	if(query_box_width*6.0f >= curr_bbox_.getWidth()) {
	    ZRectangle query_box = new ZRectangle();
	    _query_boxes.set(dg_node_id_, query_box);
	    return;
	}
	else {
	    ZRectangle query_box = new ZRectangle(query_box_top_left_x, query_box_top_left_y, 
						 query_box_width, query_box_height);
	    query_box.setPenColor(Color.black);
	    query_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    query_box.setFillColor(Color.lightGray);
	    _query_boxes.set(dg_node_id_, query_box);
	    this.addVisualComponent(query_box);
	    
	    Point2D location = new Point2D.Double(query_box_top_left_x+query_box_width/2.0f-query_label_width/2.0f,
						  query_box_top_left_y+query_box_height/2.0f-query_label_height/2.0f);
	    query_label.setPenColor(Color.black);
	    query_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of next node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    this.addVisualComponent(query_label);
	}
    }

    private void drawStarBox(int dg_node_id_, Rectangle2D curr_bbox_, double camera_mag_) {
	int font_size = getFontSize(camera_mag_);
	Font star_label_font = new Font(null, Font.PLAIN, font_size); 
	ZLabel star_label = new ZLabel(new String("*"), star_label_font);
	ZBounds star_label_bounds = star_label.getBounds();
	double star_label_width = star_label_bounds.getWidth();
	double star_label_height = star_label_bounds.getHeight();

	double star_box_width = star_label_width*3;
	double star_box_height = star_label_height; 
	double star_box_top_left_x = curr_bbox_.getX()+curr_bbox_.getWidth()*4.0f/5.0f-star_box_width/2.0f;
	double star_box_top_left_y = curr_bbox_.getY()+curr_bbox_.getHeight();
	if(star_box_width*6.0f >= curr_bbox_.getWidth()) {
	    ZRectangle star_box = new ZRectangle();
	    _star_boxes.set(dg_node_id_, star_box);
	    return;
	}
	else {
	    ZRectangle star_box = new ZRectangle(star_box_top_left_x, star_box_top_left_y, 
						 star_box_width, star_box_height);
	    star_box.setPenColor(Color.black);
	    star_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    star_box.setFillColor(Color.lightGray);
	    _star_boxes.set(dg_node_id_, star_box);
	    this.addVisualComponent(star_box);
	    
	    Point2D location = new Point2D.Double(star_box_top_left_x+star_box_width/2.0f-star_label_width/2.0f,
						  star_box_top_left_y+star_box_height/2.0f-star_label_height/2.0f);
	    star_label.setPenColor(Color.black);
	    star_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of next node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    this.addVisualComponent(star_label);
	}
    }

    private void drawPlusBox(int dg_node_id_, Rectangle2D curr_bbox_, double camera_mag_) {
	int font_size = getFontSize(camera_mag_);
	Font plus_label_font = new Font(null, Font.PLAIN, font_size); 
	ZLabel plus_label = new ZLabel(new String("+"), plus_label_font);
	ZBounds plus_label_bounds = plus_label.getBounds();
	double plus_label_width = plus_label_bounds.getWidth();
	double plus_label_height = plus_label_bounds.getHeight();

	double plus_box_width = plus_label_width*2;
	double plus_box_height = plus_label_height; 
	double plus_box_top_left_x = curr_bbox_.getX()+curr_bbox_.getWidth()*3.0f/5.0f-plus_box_width/2.0f;
	double plus_box_top_left_y = curr_bbox_.getY()+curr_bbox_.getHeight();
	if(plus_box_width*6.0f >= curr_bbox_.getWidth()) {
	    ZRectangle plus_box = new ZRectangle();
	    _plus_boxes.set(dg_node_id_, plus_box);
	    return;
	}
	else {
	    ZRectangle plus_box = new ZRectangle(plus_box_top_left_x, plus_box_top_left_y, 
						 plus_box_width, plus_box_height);
	    plus_box.setPenColor(Color.black);
	    plus_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    plus_box.setFillColor(Color.lightGray);
	    _plus_boxes.set(dg_node_id_, plus_box);
	    this.addVisualComponent(plus_box);
	    
	    Point2D location = new Point2D.Double(plus_box_top_left_x+plus_box_width/2.0f-plus_label_width/2.0f,
						  plus_box_top_left_y+plus_box_height/2.0f-plus_label_height/2.0f);
	    plus_label.setPenColor(Color.black);
	    plus_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of next node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    this.addVisualComponent(plus_label);
	}
    }

    private void drawMinusBox(int dg_node_id_, Rectangle2D curr_bbox_, double camera_mag_) {
	int font_size = getFontSize(camera_mag_);
	Font minus_label_font = new Font(null, Font.PLAIN, font_size); 
	ZLabel minus_label = new ZLabel(new String("+"), minus_label_font);
	ZBounds minus_label_bounds = minus_label.getBounds();
	double minus_label_width = minus_label_bounds.getWidth();
	double minus_label_height = minus_label_bounds.getHeight();

	double minus_box_width = minus_label_width*2;
	double minus_box_height = minus_label_height; 
	double minus_box_top_left_x = curr_bbox_.getX()+curr_bbox_.getWidth()*2.0f/5.0f-minus_box_width/2.0f;
	double minus_box_top_left_y = curr_bbox_.getY()+curr_bbox_.getHeight();
	if(minus_box_width*4.0f >= curr_bbox_.getWidth()) {
	    ZRectangle minus_box = new ZRectangle();
	    _minus_boxes.set(dg_node_id_, minus_box);
	    return;
	}
	else {
	    ZRectangle minus_box = new ZRectangle(minus_box_top_left_x, minus_box_top_left_y, 
						  minus_box_width, minus_box_height);
	    minus_box.setPenColor(Color.black);
	    minus_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    minus_box.setFillColor(Color.lightGray);
	    _minus_boxes.set(dg_node_id_, minus_box);
	    this.addVisualComponent(minus_box);
	    
	    Point2D location = new Point2D.Double(minus_box_top_left_x+minus_box_width/2.0f-minus_label_width/2.0f,
						  minus_box_top_left_y+minus_box_height/2.0f-minus_label_height/2.0f);
	    minus_label.setPenColor(Color.black);
	    minus_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of next node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    minus_label.setText(new String("-"));
	    this.addVisualComponent(minus_label);
	}
    }
    private void updateDGNodePositions() {
	findNodeTopX(_dataguide.getRoot(), 0);
	
	int curr_depth = 0;
	Vector dg_nodes_at_depth = _dataguide.getNodesAtDepth(curr_depth);
	int num_dg_nodes_at_depth = dg_nodes_at_depth.size();
	double curr_top_left_y = 0;
	while(num_dg_nodes_at_depth > 0) {
	    double max_level_height = 0;
	    for(int i=0; i<num_dg_nodes_at_depth; i++) {
		DataguideNode curr_dg_node = (DataguideNode)(dg_nodes_at_depth.elementAt(i));
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));
		if(curr_dg_rect.getHeight() > max_level_height) max_level_height = curr_dg_rect.getHeight();
	    }

	    // set the coords of the left top corner of each rectangle
	    for(int i=0; i<num_dg_nodes_at_depth; i++) {
		DataguideNode curr_dg_node = (DataguideNode)(dg_nodes_at_depth.elementAt(i));
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));		
		double curr_rect_h = curr_dg_rect.getHeight();
		double curr_rect_y = curr_top_left_y+max_level_height/2.0f-curr_rect_h/2.0f;
		curr_dg_rect.setRect(curr_dg_rect.getX(), curr_rect_y, curr_dg_rect.getWidth(), curr_dg_rect.getHeight());
	    }
	    curr_top_left_y += (max_level_height+INTER_NODE_SPACING);	    
	    curr_depth = curr_depth+1;
	    dg_nodes_at_depth = _dataguide.getNodesAtDepth(curr_depth);
	    num_dg_nodes_at_depth = dg_nodes_at_depth.size();	    
	}

	// update the zrectangles
	int num_dg_nodes = _dg_nodes.size();
	for(int i=0; i<num_dg_nodes; i++) {
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(i));	
	    ZRectangle curr_dg_zrect = (ZRectangle)(_dg_zrectangles.get(i));	
	    curr_dg_zrect.setRect(curr_dg_rect);
	}

    }

    /*
    private void updateDGNodePositions() {
	int curr_depth = 0;
	Vector dg_nodes_at_depth = _dataguide.getNodesAtDepth(curr_depth);
	int num_dg_nodes_at_depth = dg_nodes_at_depth.size();
	double curr_top_left_y = 0;
	while(num_dg_nodes_at_depth > 0) {
	    // compute the width of the current level
	    double curr_level_width = 0;
	    double max_level_height = 0;
	    for(int i=0; i<num_dg_nodes_at_depth; i++) {
		if(i != 0) curr_level_width += INTER_NODE_SPACING;
		DataguideNode curr_dg_node = (DataguideNode)(dg_nodes_at_depth.elementAt(i));
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));
		curr_level_width += curr_dg_rect.getWidth();
		if(curr_dg_rect.getHeight() > max_level_height) max_level_height = curr_dg_rect.getHeight();
	    }

	    // set the coords of the left top corner of each rectangle
	    double curr_top_left_x = -curr_level_width/2.0f;
	    for(int i=0; i<num_dg_nodes_at_depth; i++) {
		DataguideNode curr_dg_node = (DataguideNode)(dg_nodes_at_depth.elementAt(i));
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));		
		double curr_rect_w = curr_dg_rect.getWidth();
		double curr_rect_h = curr_dg_rect.getHeight();
		double curr_rect_x = curr_top_left_x;
		double curr_rect_y = curr_top_left_y+max_level_height/2.0f-curr_rect_h/2.0f;
		curr_dg_rect.setRect(curr_rect_x, curr_rect_y, curr_rect_w, curr_rect_h);
		curr_top_left_x += (curr_rect_w + INTER_NODE_SPACING);
	    }
	    curr_top_left_y += (max_level_height+INTER_NODE_SPACING);	    
	    curr_depth = curr_depth+1;
	    dg_nodes_at_depth = _dataguide.getNodesAtDepth(curr_depth);
	    num_dg_nodes_at_depth = dg_nodes_at_depth.size();	    
	}

	// update the zrectangles
	int num_dg_nodes = _dg_nodes.size();
	for(int i=0; i<num_dg_nodes; i++) {
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(i));	
	    ZRectangle curr_dg_zrect = (ZRectangle)(_dg_zrectangles.get(i));	
	    curr_dg_zrect.setRect(curr_dg_rect);
	}
    }

    */

    /**
     * function that determines the exact location of a click.
     *
     * @param click_point_ location of point where mouse was clicked
     * @param shift_down_ was the shift button down when mouse was clicked?
     */
    // find if the click point is within either the `next' or the `previous' boxes
    // If yes, create the appropriate group nodes
    synchronized public void handleClick(Point2D click_point_, boolean shift_down_) {
	double click_point_x = click_point_.getX();
	double click_point_y = click_point_.getY();
	boolean data_node_clicked = false;
	boolean prev_box_clicked = false;
	boolean next_box_clicked = false;
	boolean plus_box_clicked = false;
	boolean star_box_clicked = false;
	boolean query_box_clicked = false;
	boolean minus_box_clicked = false;

	// find which dataguide node (or equivalently group node) this click is on
	int clicked_node_num = -1;
	int num_dg_nodes = _dg_nodes.size();
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(i));	
	    if(curr_dg_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		data_node_clicked = true;
	    }
	}
		
	// check if a prev box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_prev_box_rect = ((ZRectangle)(_prev_boxes.get(i))).getRect();
	    if(curr_prev_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		prev_box_clicked = true;
	    }
	}

	// check if a next box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_next_box_rect = ((ZRectangle)(_next_boxes.get(i))).getRect();
	    if(curr_next_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		next_box_clicked = true;
	    }
	}
	
	// check if an enlarge box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_plus_box_rect = ((ZRectangle)(_plus_boxes.get(i))).getRect();
	    if(curr_plus_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		plus_box_clicked = true;
		break;
	    }
	}
	 
	// check if a toggle box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_star_box_rect = ((ZRectangle)(_star_boxes.get(i))).getRect();
	    if(curr_star_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		star_box_clicked = true;
		break;
	    }
	}
	 
	// check if a query box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_query_box_rect = ((ZRectangle)(_query_boxes.get(i))).getRect();
	    if(curr_query_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		query_box_clicked = true;
		break;
	    }
	}
	 
	// check if a shrink box was clicked
	for(int i=0; clicked_node_num == -1 && i<num_dg_nodes; i++) {
	    Rectangle2D curr_minus_box_rect = ((ZRectangle)(_minus_boxes.get(i))).getRect();
	    if(curr_minus_box_rect.contains(click_point_x, click_point_y)) {
		clicked_node_num = i;
		minus_box_clicked = true;
		break;
	    }
	}
	 
	// clicked on whitespace
	if(clicked_node_num == -1) {
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Clicked on whitespace");
	    }
	    return;
	}

	if(next_box_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Clicked on next of node "+clicked_dg_node.getLabel());
	    }
	    Integer curr_id = (Integer)(_dg_nodelistpos.get(clicked_node_num));
	    int curr_id_int = curr_id.intValue();
	    _dg_nodelistpos.set(clicked_node_num, new Integer(curr_id_int+1));
	    
	    // need to render this node
	    _repaint_flag = true;
	    this.repaint();
	}

	if(prev_box_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));		
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Clicked on prev of node "+clicked_dg_node.getLabel());
	    }
	    Integer curr_id = (Integer)(_dg_nodelistpos.get(clicked_node_num));
	    int curr_id_int = curr_id.intValue();
	    _dg_nodelistpos.set(clicked_node_num, new Integer(curr_id_int-1));
	    
	    // need to render this node
	    _repaint_flag = true;
	    this.repaint();
	}
	
	if(plus_box_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));		
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Clicked on plus of node "+clicked_dg_node.getLabel());
	    }
	    if(!shift_down_) {
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(clicked_node_num));
		double curr_pos_x = curr_dg_rect.getX();
		double curr_pos_y = curr_dg_rect.getY();
		double curr_width = curr_dg_rect.getWidth();
		double curr_height = curr_dg_rect.getHeight();
		curr_dg_rect.setRect(curr_pos_x, curr_pos_y, curr_width*ENLARGE_FACTOR, curr_height*ENLARGE_FACTOR);
		updateDGNodePositions();
	    }
	    if(shift_down_) {
		// Do I need to enlarge 
		FileVector node_set = clicked_dg_node.getNodeSet();
		int groupsize = ((Integer)_dg_nodegroupsize.get(clicked_node_num)).intValue();
		if(node_set.size() <= groupsize*groupsize) {
		    // no need to enlarge
		}
		else {
		    // enlarge
		    if(MyFrame._debug_mode > 0) {
			System.out.println("Increase Group Size");
		    }
		    groupsize++;
		    _dg_nodegroupsize.set(clicked_node_num, new Integer(groupsize));
		}
	    }
	    _repaint_flag = true;
	    this.repaint();
	}

	if(star_box_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Clicked on toggle of node "+clicked_dg_node.getLabel());
		System.out.println("Clicked on toggle of node "+clicked_dg_node.getPath());
		System.out.println("Clicked on toggle of node "+clicked_dg_node.getDGIndex());
	    }
	    boolean state = ((Boolean)(_toggled_nodes.get(clicked_node_num))).booleanValue();
	    if(state) _toggled_nodes.set(clicked_node_num, new Boolean(false));
	    else _toggled_nodes.set(clicked_node_num, new Boolean(true));

	    _repaint_flag = true;
	    this.repaint();
	}

	if(query_box_clicked) {
	    int query_node_num = clicked_node_num;
	    
	    if(_running_query != null && query_node_num == _running_query.getNode()) {
		// stop the running query
		_running_query = null;
	    }
	    else {
		String query_string = JOptionPane.showInputDialog(null, new String("Input a regular expression"), "grep", JOptionPane.QUESTION_MESSAGE);
		Regex regex = new Regex(query_string);
		DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(query_node_num));
		FileVector offsets_vector = curr_dg_node.getNodeSet();	    
		String temp_file = _datagraph.getIndexDirectory()+"partial_result";
		_running_query = new RunningQuery(_datagraph, temp_file, regex, query_node_num, offsets_vector);
		_running_query.runQuery();
		if(!_running_query.processingComplete()) _timer_thread.queueEvent(new String("RunQuery"));
	    }
	    
	    _repaint_flag = true;
	    this.repaint();
	}

	if(minus_box_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));		
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Clicked on minus of node "+clicked_dg_node.getLabel());
		System.out.println("Clicked on minus of node "+clicked_dg_node.getPath());	    
		System.out.println("Clicked on minus of node "+clicked_dg_node.getDGIndex());
	    }
	    if(!shift_down_) {
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(clicked_node_num));
		double curr_pos_x = curr_dg_rect.getX();
		double curr_pos_y = curr_dg_rect.getY();
		double curr_width = curr_dg_rect.getWidth();
		double curr_height = curr_dg_rect.getHeight();
		curr_dg_rect.setRect(curr_pos_x, curr_pos_y, curr_width/SHRINK_FACTOR, curr_height/SHRINK_FACTOR);
		updateDGNodePositions();
	    }
	    if(shift_down_) {
		if(MyFrame._debug_mode > 0) {
		    System.out.println("Decrease Group Size");
		}
		int groupsize = ((Integer)_dg_nodegroupsize.get(clicked_node_num)).intValue() - 1;
		if(groupsize < 1) groupsize = 1;
		_dg_nodegroupsize.set(clicked_node_num, new Integer(groupsize));
	    }
	    _repaint_flag = true;
	    this.repaint();
	}

	if(data_node_clicked) {
	    DataguideNode clicked_dg_node = (DataguideNode)(_dg_nodes.get(clicked_node_num));
	    String clicked_dg_path = clicked_dg_node.getPath();
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Clicked Node Label="+clicked_dg_node.getLabel());
	    }
	    
	    // find which data node was clicked
	    int num_data_nodes = _data_boxes.size();
	    for(int i=0; i<num_data_nodes; i++) {
		ZRectangle curr_bbox = (ZRectangle)(_data_boxes.get(i));
		
		if(curr_bbox.getBounds().contains(click_point_x, click_point_y)) {
		    Long clicked_offset = (Long)(_bbox_nodeoffset_map.get(curr_bbox));		    
		    if(MyFrame._debug_mode > 8) {
			DataNode clicked_data_node = _datagraph.getDataNodeNoChildren(clicked_offset.longValue());
			System.out.println("Clicked Data Node, Text="+clicked_data_node.getData());
		    }
		    // if the clicked node is already marked, unmark all nodes
		    if(_click_active && _clicked_offset == clicked_offset.longValue()) {
			_click_active = false;
		    }
		    else {
			_click_active = true;
			_clicked_offset = clicked_offset.longValue();

			// adjust the position of the list for each group node
			
			for(int j=0; j<num_dg_nodes; j++) {
			    boolean possibly_marked1 = false;
			    boolean possibly_marked2 = false;
			    DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(j));
			    String curr_dg_path = curr_dg_node.getPath();
			    if(!clicked_dg_path.equals(curr_dg_path)) {
				if(clicked_dg_path.startsWith(curr_dg_path)) {
				    System.out.println("(possible anc) : curr path="+curr_dg_path);
				    possibly_marked1 = true;  
				}
				else if(curr_dg_path.startsWith(clicked_dg_path)) {
				    System.out.println("(possible des) : curr path="+curr_dg_path);
				    possibly_marked2 = true; 
				}
			    }
			    if(possibly_marked1 || possibly_marked2) {
				FileVector curr_offsets = curr_dg_node.getNodeSet();
				int data_nodes_size = curr_offsets.size();
				int list_count = 0;
				boolean found_marked_node = false;
				long start_time = System.currentTimeMillis();
				System.out.println("Scanning list: Begin");
				curr_offsets.beginScan();
				while(list_count < data_nodes_size) {
				    boolean is_marked = false;				    
				    Long curr_offset = curr_offsets.nextElement();
				    if(possibly_marked1) {
					is_marked = _datagraph.is_ancestor(curr_offset.longValue(), _clicked_offset);
				    }
				    else if(possibly_marked2) {
					is_marked = _datagraph.is_ancestor(_clicked_offset, curr_offset.longValue());					
				    }
				    if(is_marked) {
					found_marked_node = true;
					break;
				    }
				    list_count++;		
				}
				curr_offsets.endScan();
				long end_time = System.currentTimeMillis();
				System.out.println("End of scan: list_count = "+list_count+"; time="+(end_time-start_time)+" milliseconds");
				if(found_marked_node) {
				    int groupsize = ((Integer)_dg_nodegroupsize.get(j)).intValue();
				    Integer listpos = new Integer(list_count/(groupsize*groupsize));
				    _dg_nodelistpos.set(j, listpos);
				}
			    }
			}
		    }
		    // need to render this node
		    _repaint_flag = true;
		    this.repaint();
		    break;
		}		
	    }
	}
    }
    
    /**
     * function to call repaint when timer expires at high input rate.
     */
    synchronized public void timer_expired(Object timer_event_) {    
	String object_name = timer_event_.getClass().getName();
	if(object_name.equals(new String("Histogram"))) {
	    Histogram h = (Histogram)timer_event_;
	    DataguideNode dgnode = (DataguideNode)(h.getDGNode());
	    //System.out.println("_dataguide.getNodeIndex(dgnode) = "+_dataguide.getNodeIndex(dgnode));
	    int index = dgnode.getDGIndex();
	    //System.out.println("Update check for node "+index);
	    boolean summary = ((Boolean)(_toggled_nodes.get(index))).booleanValue();
	    if(summary) {
		FileVector offsets_vector = h.getDGNode().getNodeSet();
		h.index(_datagraph, offsets_vector, h.getNextBatchNumber());
	    }
	    else { 
		//System.out.println("Histogram update ignored");
	    }
	    if(!h.processingComplete()) _timer_thread.queueEvent(h);	    
	}
	
	if(object_name.equals(new String("java.lang.String"))) { // running query
	    if(_running_query != null) { // if running query has not been stopped
		_running_query.runQuery();
		if(!_running_query.processingComplete()) _timer_thread.queueEvent(new String("RunningQuery"));
	    }
	}
	
	_repaint_flag = true;
	this.repaint();
    }

    /**
     * render function.
     * 
     * @param render_context_ context for the current render
     */
    synchronized public void render(ZRenderContext render_context_) {
	// compute magnification of camera
	double camera_mag = render_context_.getCameraMagnification();
	if(MyFrame._debug_mode > 7) {
	    System.out.println(_counter + ". ------------------------------------------");
	    System.out.println("Camera mag="+camera_mag);
	}
	double level = (camera_mag*CAMERA_MAG_SCALE_FACTOR);
	
	// get bounds of viewing window
	ZBounds viewing_window = render_context_.getVisibleBounds();
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Viewing Window. (x,y)=("+viewing_window.getBounds().getX()+","+viewing_window.getBounds().getY()+")");
	    System.out.println("Viewing Window. (w,h)=("+viewing_window.getBounds().getWidth()+","+viewing_window.getBounds().getHeight()+")");
	}

	// find dataguide node rectangles that intersect or are contained in the viewing window
	Vector visible_node_indices = new Vector();
	int num_dg_nodes = _dg_rectangles.size();
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Number of dataguide nodes="+num_dg_nodes);
	}
	for(int i=0; i<num_dg_nodes; i++) {
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(i));
	    if(viewing_window.intersects(curr_dg_rect.getX(), curr_dg_rect.getY(),
					 curr_dg_rect.getWidth(), curr_dg_rect.getHeight())) {
		visible_node_indices.addElement(new Integer(i));
	    }
	}	
	visible_node_indices.trimToSize();
	int num_visible_dg_nodes = visible_node_indices.size();
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Number of visible dataguide nodes="+num_visible_dg_nodes);
	}
	    
	// find out if panning has occurred. If yes, create new visual components.
	
	boolean panning_done = false;
	boolean zooming_done = false;
	boolean first_case = false;
	double level_diff = _curr_mag_level - level;
	if(level_diff < 0) level_diff = -level_diff;

	if(_curr_mag_level != -1 && level_diff <= MAG_CHANGE_IGNORE) {
	    if(num_visible_dg_nodes == _num_visible_dg_nodes) {
		for(int i=0; !panning_done && i<_num_visible_dg_nodes; i++) {
		    if(((Integer) visible_node_indices.elementAt(i)).intValue() !=
		       ((Integer) _visible_node_indices.elementAt(i)).intValue()) {
			panning_done = true;
			_repaint_flag = true;
		    }
		}				
	    }
	    else {
		panning_done = true;
		_repaint_flag = true;
	    }
	}
	else if(_curr_mag_level != -1) {
	    zooming_done = true;
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Zooming, time = " + System.currentTimeMillis());
	    }
	    _repaint_flag = true;
	}
	else {	 
	    first_case = true;
	    _repaint_flag = true;
	}

	if(!_repaint_flag) {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("No change, time = " + System.currentTimeMillis());
	    }
	    super.render(render_context_);
	    _counter++;
	    return;
	}
	_counter++;
	if(first_case) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Initial case, time = " + System.currentTimeMillis());
	    }
	}
	if(panning_done) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Panning, time = " + System.currentTimeMillis());
	    }
	}
	if(zooming_done) {
	    // ask the timer thread if it is time to zoom
	    // if yes, perform the zooming & reset the time_to_zoom flag in the timer thread
	    // else, queue the zoom event
	    if(_timer_thread.isTime() == false) {
		super.render(render_context_);
		if(MyFrame._debug_mode > 0) {
		    System.out.println("Queueing zoom event");
		}
		_timer_thread.queueEvent(new Long(System.currentTimeMillis()));
		return;
	    }
	}
	if(_timer_thread.isTime() == true) {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("Performing queued zoom events");
	    }
	    _timer_thread.resetTimeFlag();
	}	

	_repaint_flag = false;
	_curr_mag_level = level;
	_visible_node_indices = visible_node_indices;
	_num_visible_dg_nodes = num_visible_dg_nodes;

	super.clearVisualComponents();
	_sticky_node.clearVisualComponents();

	if(MyFrame._debug_mode > 7) {
	    System.out.println("Garbage collection begin");
	}
	System.gc();
	if(MyFrame._debug_mode > 7) {
	    System.out.println("Garbage collection end");
	}
	
		
	// compute average character width
	String sample_label = "WGglMyBm#/";
	int sample_label_len = sample_label.length();
	int font_size = getFontSize(camera_mag);
	Font sample_label_font = new Font(null, Font.PLAIN, font_size);
	ZLabel sample_label_obj = new ZLabel(sample_label, sample_label_font);
	sample_label_obj.setPenColor(Color.black);
	
	ZBounds sample_label_bounds = sample_label_obj.getBounds();
	double sample_label_width = sample_label_bounds.getWidth();
	double sample_label_height = sample_label_bounds.getHeight();	    
	double avg_char_width = sample_label_width/sample_label_len;
	double avg_char_height = sample_label_height;	    
	
	// add zrectangles to this scenegraph object
	for(int i=0; i<num_visible_dg_nodes; i++) {
	    Integer visible_node_index = (Integer)(_visible_node_indices.elementAt(i));
	    if(MyFrame._debug_mode > 7) {
		System.out.println("Visible node index="+visible_node_index);
	    }
	    //ZPolyline curr_dg_zrect = (ZPolyline)(_dg_zrectangles.get(visible_node_index.intValue()));
	    ZRectangle curr_dg_zrect = (ZRectangle)(_dg_zrectangles.get(visible_node_index.intValue()));
	    curr_dg_zrect.setPenWidth(ABS_PEN_WIDTH/camera_mag);
	    super.addVisualComponent(curr_dg_zrect);
	}
	
	// add edges between dg nodes 
	
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(i));
	    Rectangle2D curr_dg_rect = (Rectangle2D)(_dgnode_rect_map.get(curr_dg_node));

	    Vector out_edges_list = _dataguide.getOutEdges(curr_dg_node);
	    int num_out_edges = out_edges_list.size();
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Number of outgoing edges for dataguide node = " + num_out_edges);
	    }
	    if(num_out_edges == 0) continue;

	    double nearest_child_y = 0;
	    double right_most_child_center_x = 0;
	    double left_most_child_center_x = 0;
	    boolean first_time = true;
	    for(int j=0; j<num_out_edges; j++) {
		DataguideNode dest_dg_node = (DataguideNode)out_edges_list.elementAt(j);
		if(dest_dg_node == null) {
		    if(MyFrame._debug_mode > 8) {
			System.out.println("dest_dg_node is null");
		    }
		    
		}
		Rectangle2D dest_dg_rect = (Rectangle2D)_dgnode_rect_map.get(dest_dg_node);
		if(dest_dg_rect == null) {
		    if(MyFrame._debug_mode > 8) {
			System.out.println("dest_dg_rect is null");
		    }
		}
		if(first_time) {
		    first_time = false;
		    nearest_child_y = dest_dg_rect.getY();
		    left_most_child_center_x = dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0;
		    right_most_child_center_x = dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0;
		}
		else {
		    if(dest_dg_rect.getY() < nearest_child_y) {
			nearest_child_y = dest_dg_rect.getY();
		    }	
		    if((dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0) < left_most_child_center_x) {
			left_most_child_center_x = dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0;
		    }
		    if((dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0) > right_most_child_center_x) {
			right_most_child_center_x = dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0;
		    }
		}	    
	    }
	    
	    double mid_point_y = (nearest_child_y+curr_dg_rect.getY()+curr_dg_rect.getHeight())/2.0f;
	    // parent connector
	    ZPolyline parent_conn = new ZPolyline();
	    parent_conn.add(curr_dg_rect.getX()+curr_dg_rect.getWidth()/2.0f, curr_dg_rect.getY()+curr_dg_rect.getHeight());
	    parent_conn.add(curr_dg_rect.getX()+curr_dg_rect.getWidth()/2.0f, mid_point_y);
	    parent_conn.setPenColor(Color.black);
	    parent_conn.setPenWidth(ABS_PEN_WIDTH/camera_mag);
	    this.addVisualComponent(parent_conn);

	    // horizontal line
	    ZPolyline hor_conn = new ZPolyline();
	    hor_conn.add(left_most_child_center_x, mid_point_y);
	    hor_conn.add(right_most_child_center_x, mid_point_y);
	    hor_conn.setPenColor(Color.black);
	    hor_conn.setPenWidth(ABS_PEN_WIDTH/camera_mag);
	    this.addVisualComponent(hor_conn);

	    // child connectors
	    for(int j=0; j<num_out_edges; j++) {
		DataguideNode dest_dg_node = (DataguideNode)out_edges_list.elementAt(j);
		Rectangle2D dest_dg_rect = (Rectangle2D)_dgnode_rect_map.get(dest_dg_node);
		ZPolyline child_conn = new ZPolyline();
		child_conn.add(dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0f, mid_point_y);
		child_conn.add(dest_dg_rect.getX()+dest_dg_rect.getWidth()/2.0f, dest_dg_rect.getY());
		child_conn.setPenColor(Color.black);
		child_conn.setPenWidth(ABS_PEN_WIDTH/camera_mag);
		this.addVisualComponent(child_conn);
	    }
	}	

	for(int i=0; i<num_dg_nodes; i++) {
	    _prev_boxes.set(i, new ZRectangle());
	    _next_boxes.set(i, new ZRectangle());
	    _plus_boxes.set(i, new ZRectangle());
	    _minus_boxes.set(i, new ZRectangle());
	    _star_boxes.set(i, new ZRectangle());
	    _query_boxes.set(i, new ZRectangle());
	}
	_data_boxes.clear();
	_bbox_nodeoffset_map.clear();
	    
	// at low magnification, display dataguide nodes
	if(level < DG_GROUP_SWITCH_CAMERA_MAGNIFICATION) {
	    // display dataguide node labels
	    for(int i=0; i<num_visible_dg_nodes; i++) {
		Integer visible_node_index = (Integer)(_visible_node_indices.elementAt(i));
		if(MyFrame._debug_mode > 7) {
		    System.out.println("Visible node index="+visible_node_index);
		}
		DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(visible_node_index.intValue()));
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(visible_node_index.intValue()));
		String curr_label_string = curr_dg_node.getLabel();
		int curr_label_len = curr_label_string.length();
		StringBuffer new_label_buffer = new StringBuffer(" ");

		int num_chars_per_line = (int)((curr_dg_rect.getWidth()*TEXT_CORRECTION_FACTOR)/avg_char_width);
		
		if(MyFrame._debug_mode > 7) {
		    System.out.println("VDataguide: No. of chars. per line="+num_chars_per_line);
		}
		for(int j=0; j<(num_chars_per_line-1) && j<curr_label_len; j++) 
		    new_label_buffer.append(curr_label_string.charAt(j));
		String new_label_string = new_label_buffer.toString();
		if(MyFrame._debug_mode > 8) {
		    System.out.println("New Label = " + new_label_string);
                }
		ZLabel curr_label = new ZLabel(new_label_string, sample_label_font);
		ZBounds curr_label_bounds = curr_label.getBounds();
		double curr_label_width = curr_label_bounds.getWidth();///camera_mag;
		double curr_label_height = curr_label_bounds.getHeight();///camera_mag;
		Point2D location = new Point2D.Double(curr_dg_rect.getX() + curr_dg_rect.getWidth()/2.0f-curr_label_width/2.0f, 
						      curr_dg_rect.getY() + curr_dg_rect.getHeight()/2.0f-curr_label_height/2.0f);
		curr_label.setTranslation(location);
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Location of dg label in normal node coords=("+location.getX()+","+location.getY()+")");
		}
		//_my_canvas.getCamera().localToCamera(location, this);
		//curr_label.setTranslation(location);
		//if(MyFrame._debug_mode > 8) {
		//System.out.println("Location of dg label in camera coords=("+location.getX()+","+location.getY()+")");
		//}
		//_my_canvas.getCamera().cameraToLocal(location, _sticky_node);
		//curr_label.setTranslation(location);
		//if(MyFrame._debug_mode > 8) {
		//System.out.println("Location of dg label in sticky node coords=("+location.getX()+","+location.getY()+")");
		//}
		//_sticky_node.addVisualComponent(curr_label);
		this.addVisualComponent(curr_label);
		super.trimToSize();
	    }
	    
	    super.render(render_context_);
	    return;
	}
	if(level >= DG_GROUP_SWITCH_CAMERA_MAGNIFICATION) {
	    for(int i=0; i<num_visible_dg_nodes; i++) {
		Integer visible_node_index = (Integer)(_visible_node_indices.elementAt(i));
		if(MyFrame._debug_mode > 7) {
		    System.out.println("Visible node index="+visible_node_index);
		}
		//System.out.println("visible_node_index.intValue() = "+visible_node_index.intValue());
		DataguideNode curr_dg_node = (DataguideNode)(_dg_nodes.get(visible_node_index.intValue()));
		//System.out.println("curr_dg_node.getDGIndex() = "+curr_dg_node.getDGIndex());
		String curr_label_string = curr_dg_node.getLabel();		
		FileVector offsets_vector = curr_dg_node.getNodeSet();
		Rectangle2D curr_dg_rect = (Rectangle2D)(_dg_rectangles.get(visible_node_index.intValue()));		
		int group_size = ((Integer)(_dg_nodegroupsize.get(visible_node_index.intValue()))).intValue();
		drawPlusBox(visible_node_index.intValue(), curr_dg_rect, camera_mag);
		drawMinusBox(visible_node_index.intValue(), curr_dg_rect, camera_mag);
		drawStarBox(visible_node_index.intValue(), curr_dg_rect, camera_mag);
		drawQueryBox(visible_node_index.intValue(), curr_dg_rect, camera_mag);
		Integer curr_identifier = (Integer)(_dg_nodelistpos.get(visible_node_index.intValue()));
		boolean partial = false;
		
		// add visual components of this group node	    
		boolean state = ((Boolean)(_toggled_nodes.get(visible_node_index.intValue()))).booleanValue();
		if(_running_query == null || visible_node_index.intValue() != _running_query.getNode()) {
		    VGroupNode curr_group_node = new VGroupNode(visible_node_index.intValue(), _my_canvas, this, _sticky_node, 
								curr_dg_rect, curr_label_string, _dataguide, 
								_datagraph, offsets_vector, curr_identifier.intValue(),
								avg_char_width, avg_char_height, font_size, camera_mag, 
								group_size*group_size, _click_active, 
								_clicked_offset, curr_dg_node, _colors, state,
								_timer_thread, _dgnodeindex_histogram, partial);
		    
		}
		else {
		    partial = true;
		    FileVector partial_query_results = _running_query.getPartialResult();
		    VGroupNode curr_group_node = new VGroupNode(visible_node_index.intValue(), _my_canvas, this, _sticky_node, 
								curr_dg_rect, curr_label_string, _dataguide, 
								_datagraph, partial_query_results, curr_identifier.intValue(),
								avg_char_width, avg_char_height, font_size, camera_mag, 
								group_size*group_size, _click_active, 
								_clicked_offset, curr_dg_node, _colors, state, 
								_timer_thread, _dgnodeindex_histogram, 
								partial);	
		}
		
		super.trimToSize();
	    }
	    super.render(render_context_);
	    return;
	}
    }

    /**
     * function to add a data box to the Jazz object.
     *
     * @param data_bbox_ rectangle representing the databox
     * @param identifier for the data object represented by data_bbox_
     */
    synchronized public void addDataBox(ZRectangle data_bbox_, Long node_offset_) {
	_data_boxes.add(data_bbox_);
	_bbox_nodeoffset_map.put(data_bbox_, node_offset_);
    }

    /**
     * function to add a prev button to the Jazz object.
     *
     * @param prev_box_ rectangle representing the previous button
     * @param dg_node_id_ index of the dataguide node whose previous button is to be added
     */
    synchronized public void addPrevBox(ZRectangle prev_box_, int dg_node_id_) {
	_prev_boxes.set(dg_node_id_, prev_box_);
    }

    /**
     * function to add a next button to the Jazz object.
     *
     * @param next_box_ rectangle representing the next button
     * @param dg_node_id_ index of the dataguide node whose next button is to be added
     */
    synchronized public void addNextBox(ZRectangle next_box_, int dg_node_id_) {
	_next_boxes.set(dg_node_id_, next_box_);
    }
}
